#ifndef dvw_LRU_policy_mechanism_header
#define dvw_LRU_policy_mechanism_header
 
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <math.h>
 
#include "general.h"
/*
 
    Least Recently Used Policy Mechanism
 
    This class implements some of the behavior of the least recently
    eviction policy for all sizes of physical memory up to a
    user-provided limit.  In particular, given page numbers in sequence,
    it specifies, for each page number "touch", what position that page
    would have had  at that time.
 
    Here, the position with respect to this policy refers to an unsigned
    integer  value.  Returned, it means that, for all sizes of physical
    memory less than that value, this policy would have had a page-fault.
 
    The least-recently-used policy dictates that whenever a page must be
    faulted into physical memory, the page which is evicted is one which
    has not been not for the longest time.
 
    Because this policy is inclusive, the set of pages kept by
    this policy when the size of physical memory is N is a subset
    of those kept when the size is N+1.  This allows us to uniquely
    identify all pages which have been touched before by a position
    number.  (There are positions beyond our ability to care, and we
    lump those pages together with those which have never been seen
    before.)
 
    Restating, the position number associated with a page number
    at a particular time, indicates that the page is in physical memory
    if and only if the size of physical memory is greater than the
    position number.  We may think of these positions as being indexes
    into an array, positions in a linked list, etc.  However, it is
    important not to confuse these abstractions with the structures actually
    used to implement the mechanism.
 
    When each page is touched, it moves to the first position (no matter
    what the size of physical memory is, this page must in physical
    memory).  This policy dictates that it moves in the following way:
 
        Assume that the page being touched is page N and page N has
        position P.  (This is the value returned by record_touch
        member and it means that, "If the size of physical memory
        is P or greater, there are no page faults.")
 
        Page N moves to position 1, and the all the pages originally
        in positions 1 through P - 1 move to the next higher position.
 
 
*/
 
class LRU_policy
{
private:
 
/* node type */
 
   class LRU_node
   {
   public:
      page_number_type  page_number;
      AVL_status        page_tree_avl_status;
      LRU_node        * page_tree_left_child;
      LRU_node        * page_tree_right_child;
 
      position_type     size_of_left_pos_subtree;
      AVL_status        pos_tree_avl_status;
      LRU_node        * pos_tree_parent;
      LRU_node        * pos_tree_left_child;
      LRU_node        * pos_tree_right_child;
 
   };
 
 
/* global variables (for the routines in this class instantiation) */
 
   position_type  maximum_pos_number;
   position_type  node_count;
   LRU_node *     page_tree_root;
   LRU_node *     pos_tree_root;
 
/* basic AVL handling routines */
 
   void page_tree_rotate_right(LRU_node * & temp_page_tree_root)
   {
      LRU_node * save_root = temp_page_tree_root;
      temp_page_tree_root = temp_page_tree_root->page_tree_left_child;
      save_root->page_tree_left_child = temp_page_tree_root->page_tree_right_child;
      temp_page_tree_root->page_tree_right_child = save_root;
   };
   void page_tree_rotate_left(LRU_node * & temp_page_tree_root)
   {
      LRU_node * save_root = temp_page_tree_root;
      temp_page_tree_root = temp_page_tree_root->page_tree_right_child;
      save_root->page_tree_right_child = temp_page_tree_root->page_tree_left_child;
      temp_page_tree_root->page_tree_left_child = save_root;
   };
 
   /* the following two routines assume a valid AVL tree and
      return a valid AVL tree of the same height without "heaviness"
      on a particular side. */
 
   void page_tree_avoid_left_heavy(LRU_node * & temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_avl_status == LEFT_TALLER)
      {
         page_tree_avoid_right_heavy(temp_page_tree_root->page_tree_left_child);
         page_tree_rotate_right(temp_page_tree_root);
         if (temp_page_tree_root->page_tree_avl_status == BALANCED)
         {
            temp_page_tree_root->page_tree_avl_status = RIGHT_TALLER;
            (temp_page_tree_root->page_tree_right_child)->page_tree_avl_status = BALANCED;
         }
         else
         {
            temp_page_tree_root->page_tree_avl_status = RIGHT_TALLER;
            (temp_page_tree_root->page_tree_right_child)->page_tree_avl_status = RIGHT_TALLER;
         }
      };
   };
 
   void page_tree_avoid_right_heavy(LRU_node * & temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_avl_status == RIGHT_TALLER)
      {
         page_tree_avoid_left_heavy(temp_page_tree_root->page_tree_right_child);
         page_tree_rotate_left(temp_page_tree_root);
         if (temp_page_tree_root->page_tree_avl_status == BALANCED)
         {
            temp_page_tree_root->page_tree_avl_status = LEFT_TALLER;
            (temp_page_tree_root->page_tree_left_child)->page_tree_avl_status = BALANCED;
         }
         else
         {
            temp_page_tree_root->page_tree_avl_status = LEFT_TALLER;
            (temp_page_tree_root->page_tree_left_child)->page_tree_avl_status = LEFT_TALLER;
         }
      };
   };
 
   change_flag page_tree_left_child_increased(LRU_node * & temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_avl_status == LEFT_TALLER)
      {
         page_tree_avoid_right_heavy(temp_page_tree_root->page_tree_left_child);
         page_tree_rotate_right(temp_page_tree_root);
         if (temp_page_tree_root->page_tree_avl_status == BALANCED)
         {
            temp_page_tree_root->page_tree_avl_status = RIGHT_TALLER;
            (temp_page_tree_root->page_tree_right_child)->page_tree_avl_status = LEFT_TALLER;
            return INCREASED;
         }
         else /* (temp_page_tree_root == LEFT_TALLER) */
         {
            temp_page_tree_root->page_tree_avl_status = BALANCED;
            (temp_page_tree_root)->page_tree_right_child->page_tree_avl_status = BALANCED;
            return UNCHANGED;
         };
      }
      else if (temp_page_tree_root->page_tree_avl_status == BALANCED)
      {
         temp_page_tree_root->page_tree_avl_status = LEFT_TALLER;
         return INCREASED;
      }
      else /* (temp_page_tree_root->page_tree_avl_status == RIGHT_TALLER) */
      {
         temp_page_tree_root->page_tree_avl_status = BALANCED;
         return UNCHANGED;
      };
   };
 
   change_flag page_tree_right_child_increased(LRU_node * & temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_avl_status == RIGHT_TALLER)
      {
         page_tree_avoid_left_heavy(temp_page_tree_root->page_tree_right_child);
         page_tree_rotate_left(temp_page_tree_root);
         if (temp_page_tree_root->page_tree_avl_status == BALANCED)
         {
            temp_page_tree_root->page_tree_avl_status = LEFT_TALLER;
            (temp_page_tree_root->page_tree_left_child)->page_tree_avl_status = RIGHT_TALLER;
            return INCREASED;
         }
         else /* (temp_page_tree_root == RIGHT_TALLER) */
         {
            temp_page_tree_root->page_tree_avl_status = BALANCED;
            (temp_page_tree_root->page_tree_left_child)->page_tree_avl_status = BALANCED;
            return UNCHANGED;
         };
      }
      else if (temp_page_tree_root->page_tree_avl_status == BALANCED)
      {
         temp_page_tree_root->page_tree_avl_status = RIGHT_TALLER;
         return INCREASED;
      }
      else /* (temp_page_tree_root->page_tree_avl_status == LEFT_TALLER) */
      {
         temp_page_tree_root->page_tree_avl_status = BALANCED;
         return UNCHANGED;
      };
   };
 
/*
   change_flag insert_in_page_tree(LRU_node * & temp_page_tree_root,
                                   LRU_node * new_node)
   {
      if (temp_page_tree_root == NULL)
      {
         temp_page_tree_root = new_node;
         return INCREASED;
      }
      else
      {
         if (temp_page_tree_root->page_number > new_node->page_number)
         {
            if (insert_in_page_tree(temp_page_tree_root->page_tree_left_child,
                                    new_node) == INCREASED)
            {
               return page_tree_left_child_increased(temp_page_tree_root);
            };
         }
         else \* (temp_page_tree_root->page_number < new_node->page_number) *\
         {
            if (insert_in_page_tree(temp_page_tree_root->page_tree_right_child,
                                    new_node) == INCREASED)
            {
               return page_tree_right_child_increased(temp_page_tree_root);
            }
         };
      };
      return UNCHANGED;
   };
*/
 
   change_flag page_tree_left_child_decreased(LRU_node * & temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_avl_status == RIGHT_TALLER)
      {
         page_tree_avoid_left_heavy(temp_page_tree_root->page_tree_right_child);
         page_tree_rotate_left(temp_page_tree_root);
         if (temp_page_tree_root->page_tree_avl_status == BALANCED)
         {
            temp_page_tree_root->page_tree_avl_status = LEFT_TALLER;
            (temp_page_tree_root->page_tree_left_child)->page_tree_avl_status = RIGHT_TALLER;
            return UNCHANGED;
         }
         else /* (temp_page_tree_root == RIGHT_TALLER) */
         {
            temp_page_tree_root->page_tree_avl_status = BALANCED;
            (temp_page_tree_root->page_tree_left_child)->page_tree_avl_status = BALANCED;
            return DECREASED;
         };
      }
      else if (temp_page_tree_root->page_tree_avl_status == BALANCED)
      {
         temp_page_tree_root->page_tree_avl_status = RIGHT_TALLER;
         return UNCHANGED;
      }
      else /* (temp_page_tree_root->page_tree_avl_status == LEFT_TALLER) */
      {
         temp_page_tree_root->page_tree_avl_status = BALANCED;
         return DECREASED;
      };
   };
 
   change_flag page_tree_right_child_decreased(LRU_node * & temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_avl_status == LEFT_TALLER)
      {
         page_tree_avoid_right_heavy(temp_page_tree_root->page_tree_left_child);
         page_tree_rotate_right(temp_page_tree_root);
         if (temp_page_tree_root->page_tree_avl_status == BALANCED)
         {
            temp_page_tree_root->page_tree_avl_status = RIGHT_TALLER;
            (temp_page_tree_root->page_tree_right_child)->page_tree_avl_status = LEFT_TALLER;
            return UNCHANGED;
         }
         else /* (temp_page_tree_root == LEFT_TALLER) */
         {
            temp_page_tree_root->page_tree_avl_status = BALANCED;
            (temp_page_tree_root->page_tree_right_child)->page_tree_avl_status = BALANCED;
            return DECREASED;
         };
      }
      else if (temp_page_tree_root->page_tree_avl_status == BALANCED)
      {
         temp_page_tree_root->page_tree_avl_status = LEFT_TALLER;
         return UNCHANGED;
      }
      else /* (temp_page_tree_root->page_tree_avl_status == RIGHT_TALLER) */
      {
         temp_page_tree_root->page_tree_avl_status = BALANCED;
         return DECREASED;
      };
   };
 
/* These two routines require that the temp_page_tree_root be non-NULL.
 
   Note: variable is NOT passed by reference */
 
   LRU_node * find_parent_of_leftmost_in_page_tree(LRU_node * temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_left_child != NULL)
      {
         while ((temp_page_tree_root->page_tree_left_child)->page_tree_left_child != NULL)
         {
            temp_page_tree_root = temp_page_tree_root->page_tree_left_child;
         };
      };
      return temp_page_tree_root;
   };
 
   LRU_node * find_parent_of_rightmost_in_page_tree(LRU_node * temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_right_child != NULL)
      {
         while ((temp_page_tree_root->page_tree_right_child)->page_tree_right_child != NULL)
         {
            temp_page_tree_root = temp_page_tree_root->page_tree_right_child;
         };
      };
      return temp_page_tree_root;
   };
 
/* AVL routines to delete LRU nodes from the page tree --- minor changes
   for precise usage within this structure.
 
   Specifically, in the event that the node to be deleted has non-NULL
   children, we can't do the simpler trick of finding the leftmost
   node in the right subtree (or rightmost . . .) and just swapping
   the keys and data fields.  The reason is NOT that the address of
   the page node is contained within position tree part of the
   structure --- that's simple enough to fix; instead, it's because
   the routines calling this one might be holding on to a particular
   LRU node by address; thus, if we free or change that address . . .
 
   The method we use here to handle non-NULL children is to actually swap
   the positions (relative to the AVL tree), moving both nodes.  Nasty
   stuff.
 
   */
 
 
 
   void page_tree_transpose_left_to_delete(LRU_node * & temp_page_tree_root)
   {
      LRU_node * transpose_parent =
         find_parent_of_rightmost_in_page_tree(temp_page_tree_root->page_tree_left_child);
      if (transpose_parent->page_tree_right_child == NULL)
      {
         /* Here we want different storage for the pointer to the
            location we're going to transpose with --- it's our
            immediate left child and the tranposition will need
            a "save" location anyway */
         LRU_node * transpose = temp_page_tree_root->page_tree_left_child;
         temp_page_tree_root->page_tree_left_child = transpose->page_tree_left_child;
         transpose->page_tree_left_child = temp_page_tree_root;
         swap(transpose->page_tree_right_child, temp_page_tree_root->page_tree_right_child);
         swap(transpose->page_tree_avl_status, temp_page_tree_root->page_tree_avl_status);
         /* set temp_page_tree_root binding */
         temp_page_tree_root = transpose;
      }
      else
      {
         /* Here we can use a direct reference to save some complexity */
         LRU_node * & transpose = transpose_parent->page_tree_right_child;
         swap(transpose->page_tree_left_child, temp_page_tree_root->page_tree_left_child);
         swap(transpose->page_tree_right_child, temp_page_tree_root->page_tree_right_child);
         swap(transpose->page_tree_avl_status, temp_page_tree_root->page_tree_avl_status);
         /* set temp_page_tree_root and tranpose_parent->... binding */
         swap(transpose, temp_page_tree_root);
      };
   };
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
   void page_tree_transpose_right_to_delete(LRU_node * & temp_page_tree_root)
   {
      LRU_node * transpose_parent =
         find_parent_of_leftmost_in_page_tree(temp_page_tree_root->page_tree_right_child);
      if (transpose_parent->page_tree_left_child == NULL)
      {
         /* Here we want different storage for the pointer to the
            location we're going to transpose with --- it's our
            immediate right child and the tranposition will need
            a "save" location anyway */
         LRU_node * transpose = temp_page_tree_root->page_tree_right_child;
         temp_page_tree_root->page_tree_right_child = transpose->page_tree_right_child;
         transpose->page_tree_right_child = temp_page_tree_root;
         swap(transpose->page_tree_left_child, temp_page_tree_root->page_tree_left_child);
         swap(transpose->page_tree_avl_status, temp_page_tree_root->page_tree_avl_status);
         /* set temp_page_tree_root binding */
         temp_page_tree_root = transpose;
      }
      else
      {
         /* Here we can use a direct reference to save some complexity */
         LRU_node * & transpose = transpose_parent->page_tree_left_child;
         swap(transpose->page_tree_right_child, temp_page_tree_root->page_tree_right_child);
         swap(transpose->page_tree_left_child, temp_page_tree_root->page_tree_left_child);
         swap(transpose->page_tree_avl_status, temp_page_tree_root->page_tree_avl_status);
         /* set temp_page_tree_root and tranpose_parent->... binding */
         swap(transpose, temp_page_tree_root);
      };
   };
 
   change_flag extract_page_tree_root(LRU_node * & temp_root)
   {
      if ((temp_root->page_tree_left_child == NULL) &&
          (temp_root->page_tree_right_child == NULL))
      {
         free(temp_root);
         node_count--;
         temp_root = NULL;
         return DECREASED;
      }
      else if (temp_root->page_tree_avl_status == LEFT_TALLER)
      {
         page_tree_transpose_left_to_delete(temp_root);
         if (extract_rightmost_page_tree(temp_root->page_tree_left_child) == DECREASED)
         {
            return page_tree_left_child_decreased(temp_root);
         };
      }
      else
      {
         page_tree_transpose_right_to_delete(temp_root);
         if (extract_leftmost_page_tree(temp_root->page_tree_right_child) == DECREASED)
         {
            return page_tree_right_child_decreased(temp_root);
         };
      };
      return UNCHANGED;
   };
 
 
   change_flag extract_leftmost_page_tree(LRU_node * & temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_left_child == NULL)
      {
         return extract_page_tree_root(temp_page_tree_root);
      }
      else
      {
         if (extract_leftmost_page_tree(temp_page_tree_root->page_tree_left_child)
             == DECREASED)
         {
            return page_tree_left_child_decreased(temp_page_tree_root);
         }
         else
         {
            return UNCHANGED;
         };
      }
   };
 
   change_flag extract_rightmost_page_tree(LRU_node * & temp_page_tree_root)
   {
      if (temp_page_tree_root->page_tree_right_child == NULL)
      {
         return extract_page_tree_root(temp_page_tree_root);
      }
      else
      {
         if (extract_rightmost_page_tree(temp_page_tree_root->page_tree_right_child)
             == DECREASED)
         {
            return page_tree_right_child_decreased(temp_page_tree_root);
         }
         else
         {
            return UNCHANGED;
         };
      }
   };
 
 
   change_flag extract_page_tree(LRU_node * & temp_root, page_number_type page_number)
   {
      if (page_number == temp_root->page_number)
      {
         return extract_page_tree_root(temp_root);
      }
      else if (page_number < temp_root->page_number)
      {
         if (extract_page_tree(temp_root->page_tree_left_child, page_number) == DECREASED)
         {
            return page_tree_left_child_decreased(temp_root);
         };
      }
      else /* (page_number >temp_root->page_number) */
      {
         if (extract_page_tree(temp_root->page_tree_right_child, page_number) == DECREASED)
         {
            return page_tree_right_child_decreased(temp_root);
         };
      };
      return UNCHANGED;
   };
 
 
/* Position tree versions
 
   Although the position tree is also an AVL tree, it has slightly
   properties --- we require each node to retain a pointer to its
   parent and a count of the nodes in the immediate left subtree.
 
   We also do NOT require a key value with which to find things in the
   tree because, if it's not the rightmost or leftmost node we're
   concerned with, then we already know where the node is, having
   found it by indexing through the page tree.
 
   Thus, there are many subtle differences between these routines
   and those for page tree handlers. */
 
   void assert_pos_tree_left_parentage(LRU_node * temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_left_child != NULL)
      {
         (temp_pos_tree_root->pos_tree_left_child)->pos_tree_parent = temp_pos_tree_root;
      };
   };
 
   void assert_pos_tree_right_parentage(LRU_node * temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_right_child != NULL)
      {
         (temp_pos_tree_root->pos_tree_right_child)->pos_tree_parent = temp_pos_tree_root;
      };
   };
 
   void assert_pos_tree_parentage(LRU_node * temp_pos_tree_root)
   {
     assert_pos_tree_left_parentage(temp_pos_tree_root);
     assert_pos_tree_right_parentage(temp_pos_tree_root);
   };
 
   void pos_tree_rotate_right(LRU_node * & temp_pos_tree_root)
   {
      LRU_node * save_root = temp_pos_tree_root;
      temp_pos_tree_root = temp_pos_tree_root->pos_tree_left_child;
      save_root->pos_tree_left_child = temp_pos_tree_root->pos_tree_right_child;
      temp_pos_tree_root->pos_tree_right_child = save_root;
 
      temp_pos_tree_root->pos_tree_parent = save_root->pos_tree_parent;
      save_root->pos_tree_parent = temp_pos_tree_root;
      assert_pos_tree_left_parentage(save_root);
 
      save_root->size_of_left_pos_subtree = save_root->size_of_left_pos_subtree
                                            - temp_pos_tree_root->size_of_left_pos_subtree
                                            - 1;
 
 
   };
   void pos_tree_rotate_left(LRU_node * & temp_pos_tree_root)
   {
      LRU_node * save_root = temp_pos_tree_root;
      temp_pos_tree_root = temp_pos_tree_root->pos_tree_right_child;
      save_root->pos_tree_right_child = temp_pos_tree_root->pos_tree_left_child;
      temp_pos_tree_root->pos_tree_left_child = save_root;
 
      temp_pos_tree_root->pos_tree_parent = save_root->pos_tree_parent;
      save_root->pos_tree_parent = temp_pos_tree_root;
      assert_pos_tree_right_parentage(save_root);
 
      temp_pos_tree_root->size_of_left_pos_subtree = save_root->size_of_left_pos_subtree
                                            + temp_pos_tree_root->size_of_left_pos_subtree
                                            + 1;
   };
 
   /* the following two routines assume a valid AVL tree and
      return a valid AVL tree of the same height without "heaviness"
      on a particular side. */
 
   void pos_tree_avoid_left_heavy(LRU_node * & temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_avl_status == LEFT_TALLER)
      {
         pos_tree_avoid_right_heavy(temp_pos_tree_root->pos_tree_left_child);
         pos_tree_rotate_right(temp_pos_tree_root);
         if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
         {
            temp_pos_tree_root->pos_tree_avl_status = RIGHT_TALLER;
            (temp_pos_tree_root->pos_tree_right_child)->pos_tree_avl_status = BALANCED;
         }
         else
         {
            temp_pos_tree_root->pos_tree_avl_status = RIGHT_TALLER;
            (temp_pos_tree_root->pos_tree_right_child)->pos_tree_avl_status = RIGHT_TALLER;
         }
      };
   };
 
   void pos_tree_avoid_right_heavy(LRU_node * & temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_avl_status == RIGHT_TALLER)
      {
         pos_tree_avoid_left_heavy(temp_pos_tree_root->pos_tree_right_child);
         pos_tree_rotate_left(temp_pos_tree_root);
         if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
         {
            temp_pos_tree_root->pos_tree_avl_status = LEFT_TALLER;
            (temp_pos_tree_root->pos_tree_left_child)->pos_tree_avl_status = BALANCED;
         }
         else
         {
            temp_pos_tree_root->pos_tree_avl_status = LEFT_TALLER;
            (temp_pos_tree_root->pos_tree_left_child)->pos_tree_avl_status = LEFT_TALLER;
         }
      };
   };
 
   change_flag pos_tree_left_child_increased(LRU_node * & temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_avl_status == LEFT_TALLER)
      {
         pos_tree_avoid_right_heavy(temp_pos_tree_root->pos_tree_left_child);
         pos_tree_rotate_right(temp_pos_tree_root);
         if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
         {
            temp_pos_tree_root->pos_tree_avl_status = RIGHT_TALLER;
            (temp_pos_tree_root->pos_tree_right_child)->pos_tree_avl_status = LEFT_TALLER;
            return INCREASED;
         }
         else /* (temp_pos_tree_root == LEFT_TALLER) */
         {
            temp_pos_tree_root->pos_tree_avl_status = BALANCED;
            (temp_pos_tree_root)->pos_tree_right_child->pos_tree_avl_status = BALANCED;
            return UNCHANGED;
         };
      }
      else if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
      {
         temp_pos_tree_root->pos_tree_avl_status = LEFT_TALLER;
         return INCREASED;
      }
      else /* (temp_pos_tree_root->pos_tree_avl_status == RIGHT_TALLER) */
      {
         temp_pos_tree_root->pos_tree_avl_status = BALANCED;
         return UNCHANGED;
      };
   };
 
   change_flag pos_tree_right_child_increased(LRU_node * & temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_avl_status == RIGHT_TALLER)
      {
         pos_tree_avoid_left_heavy(temp_pos_tree_root->pos_tree_right_child);
         pos_tree_rotate_left(temp_pos_tree_root);
         if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
         {
            temp_pos_tree_root->pos_tree_avl_status = LEFT_TALLER;
            (temp_pos_tree_root->pos_tree_left_child)->pos_tree_avl_status = RIGHT_TALLER;
            return INCREASED;
         }
         else /* (temp_pos_tree_root == RIGHT_TALLER) */
         {
            temp_pos_tree_root->pos_tree_avl_status = BALANCED;
            (temp_pos_tree_root->pos_tree_left_child)->pos_tree_avl_status = BALANCED;
            return UNCHANGED;
         };
      }
      else if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
      {
         temp_pos_tree_root->pos_tree_avl_status = RIGHT_TALLER;
         return INCREASED;
      }
      else /* (temp_pos_tree_root->pos_tree_avl_status == LEFT_TALLER) */
      {
         temp_pos_tree_root->pos_tree_avl_status = BALANCED;
         return UNCHANGED;
      };
   };
 
   change_flag pos_tree_left_child_decreased(LRU_node * & temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_avl_status == RIGHT_TALLER)
      {
         pos_tree_avoid_left_heavy(temp_pos_tree_root->pos_tree_right_child);
         pos_tree_rotate_left(temp_pos_tree_root);
         if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
         {
            temp_pos_tree_root->pos_tree_avl_status = LEFT_TALLER;
            (temp_pos_tree_root->pos_tree_left_child)->pos_tree_avl_status = RIGHT_TALLER;
            return UNCHANGED;
         }
         else /* (temp_pos_tree_root == RIGHT_TALLER) */
         {
            temp_pos_tree_root->pos_tree_avl_status = BALANCED;
            (temp_pos_tree_root->pos_tree_left_child)->pos_tree_avl_status = BALANCED;
            return DECREASED;
         };
      }
      else if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
      {
         temp_pos_tree_root->pos_tree_avl_status = RIGHT_TALLER;
         return UNCHANGED;
      }
      else /* (temp_pos_tree_root->pos_tree_avl_status == LEFT_TALLER) */
      {
         temp_pos_tree_root->pos_tree_avl_status = BALANCED;
         return DECREASED;
      };
   };
 
   change_flag pos_tree_right_child_decreased(LRU_node * & temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_avl_status == LEFT_TALLER)
      {
         pos_tree_avoid_right_heavy(temp_pos_tree_root->pos_tree_left_child);
         pos_tree_rotate_right(temp_pos_tree_root);
         if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
         {
            temp_pos_tree_root->pos_tree_avl_status = RIGHT_TALLER;
            (temp_pos_tree_root->pos_tree_right_child)->pos_tree_avl_status = LEFT_TALLER;
            return UNCHANGED;
         }
         else /* (temp_pos_tree_root == LEFT_TALLER) */
         {
            temp_pos_tree_root->pos_tree_avl_status = BALANCED;
            (temp_pos_tree_root->pos_tree_right_child)->pos_tree_avl_status = BALANCED;
            return DECREASED;
         };
      }
      else if (temp_pos_tree_root->pos_tree_avl_status == BALANCED)
      {
         temp_pos_tree_root->pos_tree_avl_status = LEFT_TALLER;
         return UNCHANGED;
      }
      else /* (temp_pos_tree_root->pos_tree_avl_status == RIGHT_TALLER) */
      {
         temp_pos_tree_root->pos_tree_avl_status = BALANCED;
         return DECREASED;
      };
   };
 
/* These two routines require that the temp_pos_tree_root be non-NULL.
 
   Note also that variable is NOT passed by reference. */
 
   inline LRU_node * find_parent_of_leftmost_in_pos_tree(LRU_node * temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_left_child != NULL)
      {
         while ((temp_pos_tree_root->pos_tree_left_child)->pos_tree_left_child != NULL)
         {
            temp_pos_tree_root = temp_pos_tree_root->pos_tree_left_child;
         };
      };
      return temp_pos_tree_root;
   };
 
   LRU_node * find_parent_of_rightmost_in_pos_tree(LRU_node * temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_right_child != NULL)
      {
         while ((temp_pos_tree_root->pos_tree_right_child)->pos_tree_right_child != NULL)
         {
            temp_pos_tree_root = temp_pos_tree_root->pos_tree_right_child;
         };
      };
      return temp_pos_tree_root;
   };
 
/* AVL routines to extract LRU_node from the position tree.
 
   This is not a precise reflection of the code for extraction from the
   page number tree.  By arbitrary fiat, this structure does NOT own the
   nodes --- it can and does NOT free the pointers.  Thus, if you hold
   a pointer to an LRU node, you can ask for its extraction from the
   position tree and, afterwards, your pointer will still refer to an
   valid (alocated) address.  By contrast, if you hold a pointer to
   an LRU and you ask for its extraction from the page tree, the
   pointer you hold will no longer be valid.
 
   As indicated before with the page tree extraction, this code will
   retain pointer identity with page numbers --- hence, if you
 
   */
   void pos_tree_transpose_left_to_delete(LRU_node * & temp_pos_tree_root)
   {
      LRU_node * transpose_parent =
         find_parent_of_rightmost_in_pos_tree(temp_pos_tree_root->pos_tree_left_child);
      if (transpose_parent->pos_tree_right_child == NULL)
      {
         /* Here we want different storage for the pointer to the
            location we're going to transpose with --- it's our
            immediate left child and the tranposition will need
            a "save" location anyway */
         LRU_node * transpose = temp_pos_tree_root->pos_tree_left_child;
         temp_pos_tree_root->pos_tree_left_child = transpose->pos_tree_left_child;
         transpose->pos_tree_left_child = temp_pos_tree_root;
         swap(transpose->pos_tree_avl_status,
              temp_pos_tree_root->pos_tree_avl_status);
         swap(transpose->pos_tree_right_child,
              temp_pos_tree_root->pos_tree_right_child);
         swap(transpose->size_of_left_pos_subtree,
              temp_pos_tree_root->size_of_left_pos_subtree);
         transpose->pos_tree_parent = temp_pos_tree_root->pos_tree_parent;
         temp_pos_tree_root->pos_tree_parent = transpose;
         assert_pos_tree_left_parentage(temp_pos_tree_root);
         assert_pos_tree_right_parentage(transpose);
         /* change binding of temp_pos_tree_root */
         temp_pos_tree_root = transpose;
      }
      else
      {
         /* Here we can use a direct reference to save some complexity */
         LRU_node * & transpose = transpose_parent->pos_tree_right_child;
         swap(transpose->pos_tree_avl_status,
              temp_pos_tree_root->pos_tree_avl_status);
         swap(transpose->pos_tree_left_child,
              temp_pos_tree_root->pos_tree_left_child);
         swap(transpose->pos_tree_right_child,
              temp_pos_tree_root->pos_tree_right_child);
         swap(transpose->pos_tree_parent,
              temp_pos_tree_root->pos_tree_parent);
         swap(transpose->size_of_left_pos_subtree,
              temp_pos_tree_root->size_of_left_pos_subtree);
         assert_pos_tree_parentage(temp_pos_tree_root);
         assert_pos_tree_parentage(transpose);
         /* change binding of temp_pos_tree_root */
         swap(transpose, temp_pos_tree_root);
      };
   };
 
   void pos_tree_transpose_right_to_delete(LRU_node * & temp_pos_tree_root)
   {
      LRU_node * transpose_parent =
         find_parent_of_leftmost_in_pos_tree(temp_pos_tree_root->pos_tree_right_child);
      if (transpose_parent->pos_tree_left_child == NULL)
      {
         /* Here we want different storage for the pointer to the
            location we're going to transpose with --- it's our
            immediate right child and the tranposition will need
            a "save" location anyway */
         LRU_node * transpose = temp_pos_tree_root->pos_tree_right_child;
         temp_pos_tree_root->pos_tree_right_child = transpose->pos_tree_right_child;
         transpose->pos_tree_right_child = temp_pos_tree_root;
         swap(transpose->pos_tree_avl_status,
              temp_pos_tree_root->pos_tree_avl_status);
         swap(transpose->pos_tree_left_child,
              temp_pos_tree_root->pos_tree_left_child);
         swap(transpose->size_of_left_pos_subtree,
              temp_pos_tree_root->size_of_left_pos_subtree);
         transpose->pos_tree_parent = temp_pos_tree_root->pos_tree_parent;
         temp_pos_tree_root->pos_tree_parent = transpose;
         assert_pos_tree_right_parentage(temp_pos_tree_root);
         assert_pos_tree_left_parentage(transpose);
         /* change binding of temp_pos_tree_root */
         temp_pos_tree_root = transpose;
      }
      else
      {
         /* Here we can use a direct reference to save some complexity */
         LRU_node * & transpose = transpose_parent->pos_tree_left_child;
         swap(transpose->pos_tree_avl_status,
              temp_pos_tree_root->pos_tree_avl_status);
         swap(transpose->pos_tree_right_child,
              temp_pos_tree_root->pos_tree_right_child);
         swap(transpose->pos_tree_left_child,
              temp_pos_tree_root->pos_tree_left_child);
         swap(transpose->pos_tree_parent,
              temp_pos_tree_root->pos_tree_parent);
         swap(transpose->size_of_left_pos_subtree,
              temp_pos_tree_root->size_of_left_pos_subtree);
         assert_pos_tree_parentage(temp_pos_tree_root);
         assert_pos_tree_parentage(transpose);
         /* change binding of temp_pos_tree_root */
         swap(transpose, temp_pos_tree_root);
      };
   };
 
   change_flag extract_pos_tree_root(LRU_node * & temp_pos_tree_root)
   {
      if ((temp_pos_tree_root->pos_tree_left_child == NULL) &&
          (temp_pos_tree_root->pos_tree_right_child == NULL))
      {
         temp_pos_tree_root->pos_tree_parent = NULL;
         temp_pos_tree_root = NULL;
 
         return DECREASED;
      }
      else if (temp_pos_tree_root->pos_tree_avl_status == LEFT_TALLER)
      {
         pos_tree_transpose_left_to_delete(temp_pos_tree_root);
         temp_pos_tree_root->size_of_left_pos_subtree--;
         if (extract_rightmost_pos_tree(temp_pos_tree_root->pos_tree_left_child) == DECREASED)
         {
            return pos_tree_left_child_decreased(temp_pos_tree_root);
         };
      }
      else
      {
         pos_tree_transpose_right_to_delete(temp_pos_tree_root);
         if (extract_leftmost_pos_tree(temp_pos_tree_root->pos_tree_right_child) == DECREASED)
         {
            return pos_tree_right_child_decreased(temp_pos_tree_root);
         };
      };
      return UNCHANGED;
   };
 
   change_flag extract_leftmost_pos_tree(LRU_node * & temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_left_child == NULL)
      {
         return extract_pos_tree_root(temp_pos_tree_root);
      }
      else
      {
         temp_pos_tree_root->size_of_left_pos_subtree--;
         if (extract_leftmost_pos_tree(temp_pos_tree_root->pos_tree_left_child)
             == DECREASED)
         {
            return pos_tree_left_child_decreased(temp_pos_tree_root);
         }
         else
         {
            return UNCHANGED;
         };
      }
   };
 
   change_flag extract_rightmost_pos_tree(LRU_node * & temp_pos_tree_root)
   {
      if (temp_pos_tree_root->pos_tree_right_child == NULL)
      {
         return extract_pos_tree_root(temp_pos_tree_root);
      }
      else
      {
         if (extract_rightmost_pos_tree(temp_pos_tree_root->pos_tree_right_child)
             == DECREASED)
         {
            return pos_tree_right_child_decreased(temp_pos_tree_root);
         }
         else
         {
            return UNCHANGED;
         };
      }
   };
 
/* We've "unrolled" the usual recursion into a while loop
   here --- strictly speaking, we can do this better than the
   compiler (*insuring* no stack overhead, etc.), and probably
   should do this for other routines as well.  The reason we haven't
   done this everywhere is another minor concession to code clarity.
   Frankly, it's often to difficult to track errors in either
   a recursive version or a non-recursive version; however, it
   is less-difficult to see, understand, and track the effects of
   booleans.
 
   The immediate example here is the "test_which_side" variable.
   What it means, if true, is that the node in question was the left
   child of its parent.  Here, we have to reset it almost before we
   it, giving a difficult-to-follow set of nested if statements
   in order to get to the "right" pointer reference.
 
   The better reason for making this non-recursive is that we
   really need to do different things during the recursion (acculmulating
   the position value, etc) AND we're going in the opposite
   direction (upwards as opposed to downwards) of the methods generally
   for AVL tree traversal (in this structure and related code).
   Thus, a recursive version would be less clear as well as less
   efficient.
 
   Note that these statements will have been effectively executed as
   part of the extraction process:
 
      touched->size_of_left_pos_subtree = 0;
      touched->pos_tree_avl_status      = BALANCED;
      touched->pos_tree_parent          = NULL;
      touched->pos_tree_left_child      = NULL;
      touched->pos_tree_right_child     = NULL;
*/
 
   void calculate_position_and_extract_pos_tree(LRU_node * touched,
                                                position_type & position)
   {
      LRU_node *  parent = touched->pos_tree_parent;
 
      position   = 1 + touched->size_of_left_pos_subtree;
      if (parent == NULL)
      {
         extract_pos_tree_root(pos_tree_root);
      }
      else
      {
         change_flag change;
 
         int test_which_side = (parent->pos_tree_left_child == touched);
         if (test_which_side)
         {
            parent->size_of_left_pos_subtree--;
            change = extract_pos_tree_root(parent->pos_tree_left_child);
         }
         else
         {
            position = position + parent->size_of_left_pos_subtree + 1;
            change = extract_pos_tree_root(parent->pos_tree_right_child);
         };
 
         LRU_node * pruned = parent;
         parent = pruned->pos_tree_parent;
         while (parent != NULL)
         {
            if (test_which_side)
            {
               test_which_side = (parent->pos_tree_left_child == pruned);
               if (test_which_side)
               {
                  parent->size_of_left_pos_subtree--;
                  if (change == DECREASED)
                  {
                     change = pos_tree_left_child_decreased(parent->pos_tree_left_child);
                  };
               }
               else
               {
                  position = position + parent->size_of_left_pos_subtree + 1;
                  if (change == DECREASED)
                  {
                     change = pos_tree_left_child_decreased(parent->pos_tree_right_child);
                  };
               };
            }
            else
            {
               test_which_side = (parent->pos_tree_left_child == pruned);
               if (test_which_side)
               {
                  parent->size_of_left_pos_subtree--;
                  if (change == DECREASED)
                  {
                     change = pos_tree_right_child_decreased(parent->pos_tree_left_child);
                  };
               }
               else
               {
                  position = position + parent->size_of_left_pos_subtree + 1;
                  if (change == DECREASED)
                  {
                     change = pos_tree_right_child_decreased(parent->pos_tree_right_child);
                  };
               };
            };
            pruned = parent;
            parent = pruned->pos_tree_parent;
         };
         if (test_which_side)
         {
            if (change == DECREASED)
            {
               pos_tree_left_child_decreased(pos_tree_root);
            };
         }
         else
         {
            if (change == DECREASED)
            {
               pos_tree_right_child_decreased(pos_tree_root);
            };
         };
      };
   };
 
   void free_LRU_last()
   {
      LRU_node * condemned_node = pos_tree_root;
      while (condemned_node->pos_tree_right_child != NULL)
      {
         condemned_node = condemned_node->pos_tree_right_child;
      };
      page_number_type condemned_page_number = condemned_node->page_number;
      extract_rightmost_pos_tree(pos_tree_root);
      extract_page_tree(page_tree_root, condemned_page_number);
   };
 
 
/* moves an LRU_node into the first position */
 
   change_flag insert_left_of_pos_tree(LRU_node * & temp_pos_tree_root,
                                       LRU_node * new_node)
   {
      if (temp_pos_tree_root == NULL)
      {
         temp_pos_tree_root = new_node;
         return INCREASED;
      }
      else
      {
         temp_pos_tree_root->size_of_left_pos_subtree++;
         new_node->pos_tree_parent = temp_pos_tree_root;
         if (insert_left_of_pos_tree(temp_pos_tree_root->pos_tree_left_child,
                                     new_node) == INCREASED)
         {
            return pos_tree_left_child_increased(temp_pos_tree_root);
         };
      };
      return UNCHANGED;
   };
 
/* We're touching a page, and this routine finds it in the AVL
   tree --- the assumption is that it doesn't exist and needs
   to be added to the tree --- hence, we save the trace to the node
   (using the function stack) to avoid traversing the tree a second
   time along the same path.
 
   This intermittant inefficiency can't easily be avoided.  We can't
   save the trace into the tree for the duration of the touch operation
   becuase the touch operation may change tree in other ways --- we
   may be able to release some pages.  We can't simply attempt to
   find and create the node in one pass because we really need to do
   some different things if the node is new. */
 
 
 
   change_flag touch_through_page_tree(LRU_node * & temp_page_tree_root,
                                       page_number_type page_number,
                                       position_type & position)
   {
      if (temp_page_tree_root == NULL)
      {
         frail_allocate(temp_page_tree_root, LRU_node);
         temp_page_tree_root->page_number              = page_number;
         temp_page_tree_root->page_tree_avl_status     = BALANCED;
         temp_page_tree_root->page_tree_left_child     = NULL;
         temp_page_tree_root->page_tree_right_child    = NULL;
         temp_page_tree_root->size_of_left_pos_subtree = 0;
         temp_page_tree_root->pos_tree_avl_status      = BALANCED;
         temp_page_tree_root->pos_tree_parent          = NULL;
         temp_page_tree_root->pos_tree_left_child      = NULL;
         temp_page_tree_root->pos_tree_right_child     = NULL;
         insert_left_of_pos_tree(pos_tree_root, temp_page_tree_root);
         node_count++;
         position = maximum_pos_number;
         return INCREASED;
      }
      else
      {
         if (temp_page_tree_root->page_number > page_number)
         {
            if (touch_through_page_tree(temp_page_tree_root->page_tree_left_child,
                                        page_number, position) == INCREASED)
            {
               return page_tree_left_child_increased(temp_page_tree_root);
            };
         }
         else if (temp_page_tree_root->page_number < page_number)
         {
            if (touch_through_page_tree(temp_page_tree_root->page_tree_right_child,
                                        page_number, position) == INCREASED)
            {
               return page_tree_right_child_increased(temp_page_tree_root);
            }
         }
         else /* (temp_page_tree_root->page_number = page_number) */
         {
            calculate_position_and_extract_pos_tree(temp_page_tree_root, position);
            insert_left_of_pos_tree(pos_tree_root, temp_page_tree_root);
         }
      };
      return UNCHANGED;
   };
 
   void kill_through_page_tree(LRU_node * temp_root)
   {
      if (temp_root != NULL)
      {
         kill_through_page_tree(temp_root->page_tree_left_child);
         kill_through_page_tree(temp_root->page_tree_right_child);
         free(temp_root);
      };
   };
 
public:
   LRU_policy(position_type maximum_pos_request)
   {
      maximum_pos_number = maximum_pos_request;
      page_tree_root     = NULL;
      pos_tree_root      = NULL;
      node_count         = 0;
   };
 
   void record_touch(page_number_type page_number,
                     position_type & position)
   {
      touch_through_page_tree(page_tree_root, page_number, position);
      if (node_count > maximum_pos_number)
      {
         free_LRU_last();
      };
   };
 
   void check_page_tree_condition(LRU_node * temp_root, int & height)
   {
      if (temp_root == NULL)
      {
         height = 0;
      }
      else
      {
         int left_height;
         int right_height;
         check_page_tree_condition(temp_root->page_tree_left_child, left_height);
         check_page_tree_condition(temp_root->page_tree_right_child, right_height);
 
         int test;
         switch (temp_root->page_tree_avl_status)
         {
            case LEFT_TALLER:
            {
               test = (left_height == right_height + 1);
               height = left_height + 1;
            }; break;
            case RIGHT_TALLER:
            {
               test = (left_height + 1 == right_height);
               height = right_height + 1;
            }; break;
            case BALANCED:
            {
               test = (left_height == right_height);
               height = right_height + 1;
            }; break;
            default:
            {
               test = 0;
            }
         };
         if (!test)
         {
            fprintf(stderr, "Possible corruption in LRU mechanism\n");
            fprintf(stderr, "Terminating prematurely \n");
            exit(1);
         };
      };
 
   };
 
   void check_pos_tree_condition(LRU_node * parent, LRU_node * temp_root,
                                 int & height, int & weight)
   {
      if (temp_root == NULL)
      {
         height = 0;
         weight = 0;
      }
      else
      {
         int left_height;
         int right_height;
         int left_weight;
         int right_weight;
         check_pos_tree_condition(temp_root, temp_root->pos_tree_left_child,
                                  left_height, left_weight);
         check_pos_tree_condition(temp_root, temp_root->pos_tree_right_child,
                                  right_height, right_weight);
 
         weight = left_weight + right_weight + 1;
         int test = ((temp_root->pos_tree_parent == parent) &&
                     (temp_root->size_of_left_pos_subtree == left_weight));
         switch (temp_root->pos_tree_avl_status)
         {
            case LEFT_TALLER:
            {
               test = ((left_height == right_height + 1) && test);
               height = left_height + 1;
            }; break;
            case RIGHT_TALLER:
            {
               test = ((left_height + 1 == right_height) && test);
               height = right_height + 1;
            }; break;
            case BALANCED:
            {
               test = ((left_height == right_height) && test);
               height = right_height + 1;
            }; break;
            default:
            {
               test = 0;
            }
         };
         if (!test)
         {
            fprintf(stderr, "Possible corruption in LRU mechanism\n");
            fprintf(stderr, "Terminating prematurely \n");
            exit(1);
         };
      };
 
   };
 
   void kill_and_test_for_corruption()
   {
      int height;
      int weight;
      check_page_tree_condition(page_tree_root, height);
      check_pos_tree_condition(NULL, pos_tree_root, height, weight);
      kill_through_page_tree(page_tree_root);
      page_tree_root     = NULL;
      pos_tree_root      = NULL;
      node_count         = 0;
      fprintf(stderr, "LRU policy mechanism successfully released \n");
   };
};
 
 
#endif

