#ifndef my_basic_plot_stuff
#define my_basic_stuff
 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <float.h>
#include <math.h>
#include <memory.h>
#include "general.h"
 
/* This is explicitly landscape oriented, thus the height is 8.5
   inches.  These macro definitions are fixed in terms of 72 points
   per inch, still a standard in terms of point size, and the
   natural scale for postscript files.
*/
 
#define PIXELS_PER_INCH        (72.0)
 
#define PAGE_WIDTH_INCHES      (11)
#define PAGE_HEIGHT_INCHES     (8.5)
 
#define LEFT_MARGIN_INCHES     (0.5)
#define RIGHT_MARGIN_INCHES    (0.5)
#define TOP_MARGIN_INCHES      (0.5)
#define BOTTOM_MARGIN_INCHES   (0.5)
 
#define INTERNAL_MARGIN_INCHES (0.5)
 
#define LEGEND_SIZE_INCHES     (1.5)
 
#define PAGE_WIDTH             (PIXELS_PER_INCH * PAGE_WIDTH_INCHES)
#define PAGE_HEIGHT            (PIXELS_PER_INCH * PAGE_HEIGHT_INCHES)
#define LEFT_MARGIN            (PIXELS_PER_INCH * LEFT_MARGIN_INCHES)
#define RIGHT_MARGIN           (PIXELS_PER_INCH * RIGHT_MARGIN_INCHES)
#define TOP_MARGIN             (PIXELS_PER_INCH * TOP_MARGIN_INCHES)
#define BOTTOM_MARGIN          (PIXELS_PER_INCH * BOTTOM_MARGIN_INCHES)
#define INTERNAL_MARGIN        (PIXELS_PER_INCH * INTERNAL_MARGIN_INCHES)
#define LEGEND_SIZE            (PIXELS_PER_INCH * LEGEND_SIZE_INCHES)
 
/* The number of tics along an axis doesn't look good if there are more than
   this number --- yes, it's a guess.  */
#define MAXIMUM_MAJOR_TICS  (13)
 
/* Length of tic in pixels */
#define MAJOR_TIC_SIZE         (8)
#define MINOR_TIC_SIZE         (6)
#define VERY_MINOR_TIC_SIZE    (4)
 
/* Thickness of lines */
#define BOX_LINE_THICKNESS       (0.20)
#define MAJOR_TIC_THICKNESS      (0.20)
#define MINOR_TIC_THICKNESS      (0.15)
#define VERY_MINOR_TIC_THICKNESS (0.10)
 
/* Most reasonable fonts are rectangular, the width being a fraction of
   the height --- this is used strictly with numerals to estimate
   the worst case size of an axis label. */
 
#define FONT_WIDTH_RATIO_ESTIMATE (0.6)
 
#define VERSION_NAME ("Internal PostScript Plotter --- Version 1.3")
 
 
/* Some enum types for communicating with and within the plot class */
 
enum value_format
{
   HEXADECIMAL_NUMBERS,
   DECIMAL_NUMBERS,
   BOTH_HEX_AND_DEC
};
 
enum justification
{
   LEFT_JUSTIFY,
   RIGHT_JUSTIFY,
   CENTER
};
 
enum rotate_direction
{
   DO_NOT_ROTATE,
   ROTATE_NINETY,
   ROTATE_MINUS_NINETY
};
 
enum image_association
{
   PLOT_IMAGE,
   LEGEND_IMAGE,
   ALL_IMAGES,
   NOT_SHOWN
};
 
class plot_basics
{
/* Useful graphics junk */
protected:
   void emit_box(FILE * output_stream,
                 double left_x_position,
                 double right_x_position,
                 double upper_y_position,
                 double lower_y_position)
   {
      fprintf(output_stream, "newpath %g %g moveto %% BOX\n",
                             left_x_position, upper_y_position);
      fprintf(output_stream, "        %g %g lineto\n",
                             left_x_position, lower_y_position);
      fprintf(output_stream, "        %g %g lineto\n",
                             right_x_position, lower_y_position);
      fprintf(output_stream, "        %g %g lineto\n",
                             right_x_position, upper_y_position);
      fprintf(output_stream, "closepath %g setlinewidth stroke\n",
                             BOX_LINE_THICKNESS);
   };
 
   void emit_outline_box(FILE * output_stream,
                         double left_x_position,
                         double right_x_position,
                         double upper_y_position,
                         double lower_y_position)
   {
      emit_box(output_stream,
               left_x_position  - 0.5 * BOX_LINE_THICKNESS,
               right_x_position + 0.5 * BOX_LINE_THICKNESS,
               upper_y_position + 0.5 * BOX_LINE_THICKNESS,
               lower_y_position - 0.5 * BOX_LINE_THICKNESS);
 
   };
 
   void emit_block(FILE * output_stream,
                   double left_x_position,
                   double right_x_position,
                   double upper_y_position,
                   double lower_y_position,
                   double setgray)
   {
      fprintf(output_stream, "newpath %g %g moveto %% BLOCK\n",
                             left_x_position, upper_y_position);
      fprintf(output_stream, "        %g %g lineto\n",
                             left_x_position, lower_y_position);
      fprintf(output_stream, "        %g %g lineto\n",
                             right_x_position, lower_y_position);
      fprintf(output_stream, "        %g %g lineto\n",
                             right_x_position, upper_y_position);
 
      fprintf(output_stream, "closepath %g setgray fill\n",
                             setgray);
   };
 
   int compatible_images(image_association image1,
                         image_association image2)
   {
      return ((image1 == image2) ||
              (image1 == ALL_IMAGES) ||
              (image2 == ALL_IMAGES));
   };
 
/*
   This structure gives us a means of placing items such as plot text
   relative to one another without calculating exactly where the base
   item is first.  For example, when the user requests that the legend
   have text centered above it, we'd need to know where the legend
   was, but, until the user has exhausted all optionial requests, we
   may not be able to calculate the position exactly.
*/
public:
   struct movable_position
   {
      boolean            fixed;
      double             position;
      boolean            midpoint;
      movable_position * mp1_ptr;
      movable_position * mp2_ptr;
      double             scale;
      double             offset;
   };
 
protected:
   void set_position_fixed(movable_position & mp,
                           double position)
   {
      mp.fixed    = LOGICAL_TRUE;
      mp.position = position;
   };
 
   void set_position_at_midpoint(movable_position & mp,
                                 movable_position * address_of_position_1,
                                 movable_position * address_of_position_2)
   {
      mp.fixed    = LOGICAL_FALSE;
      mp.midpoint = LOGICAL_TRUE;
      mp.mp1_ptr  = address_of_position_1;
      mp.mp2_ptr  = address_of_position_2;
      mp.scale   = 0.5;
   };
 
   void set_position_relative_midpoint(movable_position & mp,
                                       movable_position * address_of_position_1,
                                       movable_position * address_of_position_2,
                                       double fraction)
   {
      mp.fixed    = LOGICAL_FALSE;
      mp.midpoint = LOGICAL_TRUE;
      mp.mp1_ptr  = address_of_position_1;
      mp.mp2_ptr  = address_of_position_2;
      mp.scale    = fraction;
   };
 
   void set_position_relative_to(movable_position & mp,
                                 movable_position * address_of_position,
                                 double offset)
   {
      mp.fixed    = LOGICAL_FALSE;
      mp.midpoint = LOGICAL_FALSE;
      mp.mp1_ptr  = address_of_position;
      mp.offset   = offset;
   };
 
   void set_position_link_to(movable_position & mp,
                             movable_position * address_of_position)
   {
      mp.fixed    = LOGICAL_FALSE;
      mp.midpoint = LOGICAL_FALSE;
      mp.mp1_ptr  = address_of_position;
      mp.offset   = 0.0;
   };
 
   void set_positions_overlap_link(movable_position & mp,
                                   movable_position * address_of_position_1,
                                   movable_position * address_of_position_2)
   {
      set_position_link_to(mp, address_of_position_2);
      set_position_link_to(* address_of_position_1, & mp);
   };
 
   void squeeze_before(movable_position & wedge_mp,
                       double             offset_before,
                       double             offset_after,
                       movable_position & mp)
   {
      wedge_mp         = mp;
      wedge_mp.offset += offset_before;
      mp.offset       += offset_before + offset_after;
   };
 
   void squeeze_before_overlap(movable_position & wedge_mp,
                               double             offset_before,
                               double             offset_after,
                               movable_position & mp)
   {
      wedge_mp            = mp;
      wedge_mp.offset    += offset_before;
      mp.offset          += offset_before + offset_after;
      mp.mp1_ptr->offset -= offset_before + offset_after;
   };
 
 
   double position(movable_position mp)
   {
      if (mp.fixed == LOGICAL_TRUE)
      {
         return mp.position;
      }
      else
      {
         if (mp.midpoint == LOGICAL_TRUE)
         {
            return (position(* mp.mp1_ptr) * (1.0 - mp.scale)) +
                    position(* mp.mp2_ptr) * mp.scale;
         }
         else
         {
            return mp.offset + position(* mp.mp1_ptr);
         };
      };
   };
 
   void emit_movable_position(FILE * output_stream, movable_position mp)
   {
      fprintf(output_stream, "%g", position(mp));
   };
 
   struct scale_info
   {
       UInt64              user_min;
       UInt64              user_max;
       UInt32              absolute_storage_max;
       UInt32              storage_max;
 
       movable_position *  box_loc_min_ptr;
       movable_position *  box_loc_max_ptr;
       movable_position *  box_loc_inner_ptr;
       movable_position *  box_loc_outer_ptr;
       movable_position    axis_center;
       movable_position    label_edge;
 
       double              user_2_storage;
       double              unit_fraction;
 
/* display stuff for axis and labels */
 
       boolean             tics_requested;
       boolean             minor_tics_requested;
       UInt64              first_major_tic;
       UInt64              first_minor_tic;
       UInt64              first_very_minor_tic;
       UInt64              major_spacing;
       UInt64              minor_spacing;
       UInt64              very_minor_spacing;
 
       boolean             labels_requested;
       char *              font_name;
       double              font_size;
       value_format        style;
 
       image_association   image;
   };
 
/* Where the scale "fits" in the big picture */
   void setup_box_location(scale_info & si,
                           movable_position * box_loc_min_ptr,
                           movable_position * box_loc_max_ptr,
                           movable_position * box_loc_outer_ptr,
                           movable_position * box_loc_inner_ptr)
   {
      si.box_loc_min_ptr   = box_loc_min_ptr;
      si.box_loc_max_ptr   = box_loc_max_ptr;
      si.box_loc_outer_ptr = box_loc_outer_ptr;
      si.box_loc_inner_ptr = box_loc_inner_ptr;
 
      set_position_link_to(* box_loc_inner_ptr,
                           & si.axis_center);
      set_position_link_to(si.axis_center,
                           & si.label_edge);
      set_position_link_to(si.label_edge,
                           si.box_loc_outer_ptr);
   };
 
   void verify_scale_area(scale_info & si)
   {
      if (si.tics_requested == LOGICAL_TRUE)
      {
         si.box_loc_inner_ptr->offset = MAJOR_TIC_SIZE;
         si.axis_center.offset        = 0.5 * MAJOR_TIC_SIZE;
      }
      else
      {
         si.box_loc_inner_ptr->offset = 0;
         si.axis_center.offset        = 0;
      };
      if (si.labels_requested == LOGICAL_TRUE)
      {
         si.axis_center.offset += 0.5 * si.font_size;
         si.label_edge.offset = maximum_label_size(si);
      }
      else
      {
         si.label_edge.offset = 0.0;
      };
   };
 
/* Set initial stuff */
 
   void setup_scale_parameters(scale_info & si,
                               UInt32 absolute_storage_max,
                               image_association image)
   {
      si.absolute_storage_max = absolute_storage_max;
      si.tics_requested       = LOGICAL_TRUE;
      si.minor_tics_requested = LOGICAL_TRUE;
      si.labels_requested     = LOGICAL_TRUE;
      si.font_name            = frail_strdup("Courier");
      si.font_size            = 10.0;
      si.style                = BOTH_HEX_AND_DEC;
      si.image                = image;
   };
 
/* Alter the range */
 
   void setup_range(scale_info & si, UInt64 user_min, UInt64 user_max)
   {
      si.user_min = user_min;
      si.user_max = user_max;
      UInt64 range = user_max - user_min;
 
      if (range < si.absolute_storage_max)
      {
         si.storage_max                 = range;
         si.user_2_storage              = 1.0;
      }
      else
      {
         double uupsu = ceil((range + 1.0) / si.absolute_storage_max);
         fprintf(stderr, "user_units per storage_unit %g \n", uupsu);
         si.user_2_storage = 1.0 / uupsu;
         si.storage_max = (UInt32)ceil(range * si.user_2_storage);
      };
      si.unit_fraction        = 1.0 / (range + 1.0);
   };
 
   void setup_range_force_storage_max(scale_info & si,
                                      UInt64 user_min,
                                      UInt64 user_max)
   {
      si.user_min = user_min;
      si.user_max = user_max;
      double range = user_max - user_min;
      si.storage_max = si.absolute_storage_max;
      si.user_2_storage = si.storage_max / range;
      si.unit_fraction        = 1.0 / (range + 1.0);
   };
 
 
/* Use the scale to translate from user positions to storage positions,
   from user positions to ... */
 
   UInt32 storage_location(scale_info si, UInt64 value)
   {
      return (UInt32)((value - si.user_min) * si.user_2_storage);
   };
 
   UInt32 storage_location_method_two(scale_info si, UInt64 value)
   {
      return (UInt32)ceil((value - si.user_min) * si.user_2_storage);
   };
 
   movable_position box_edge_first_from_storage(scale_info si, UInt32 value)
   {
      movable_position temp;
      set_position_relative_midpoint(temp,
                                     si.box_loc_min_ptr,
                                     si.box_loc_max_ptr,
                                     value * si.unit_fraction /
                                             si.user_2_storage);
      return temp;
   };
 
   movable_position box_edge_first(scale_info si, UInt64 value)
   {
      movable_position temp;
      set_position_relative_midpoint(temp,
                                     si.box_loc_min_ptr,
                                     si.box_loc_max_ptr,
                                     (value - si.user_min) * si.unit_fraction);
      return temp;
   };
 
   movable_position box_center(scale_info si, UInt64 value)
   {
      movable_position temp;
      set_position_relative_midpoint(temp,
                                     si.box_loc_min_ptr,
                                     si.box_loc_max_ptr,
                                     (value + 0.5 - si.user_min) * si.unit_fraction);
      return temp;
   };
 
   movable_position box_edge_second(scale_info si, UInt64 value)
   {
      movable_position temp;
      set_position_relative_midpoint(temp,
                                     si.box_loc_min_ptr,
                                     si.box_loc_max_ptr,
                                     (value + 1.0 - si.user_min) * si.unit_fraction);
      return temp;
   };

   movable_position box_edge_second_from_storage(scale_info si, UInt32 value)
   {
      movable_position temp;
      set_position_relative_midpoint(temp,
                                     si.box_loc_min_ptr,
                                     si.box_loc_max_ptr,
                                     value * si.unit_fraction / 
                                             si.user_2_storage
                                     + si.unit_fraction);
      return temp;
   };
 
 
/* Cartesian coordinate handling */
public:
   struct coordinate_pair
   {
      movable_position              x;
      movable_position              y;
   };
 
protected:
   void emit_coordinate_pair(FILE * output_stream, coordinate_pair cp)
   {
      emit_movable_position(output_stream, cp.x);
      fprintf(output_stream, " ");
      emit_movable_position(output_stream, cp.y);
   };
 
   void emit_coordinate_pair(FILE * output_stream, coordinate_pair cp,
                        const char * append_format_string)
   {
      emit_movable_position(output_stream, cp.x);
      fprintf(output_stream, " ");
      emit_movable_position(output_stream, cp.y);
      fprintf(output_stream, " ");
      fprintf(output_stream, append_format_string);
   };
 
/* The big engine which makes this work */
 
   struct plot_text
   {
      char *                  text_string;
      char *                  font_name;
      double                  font_size;
      rotate_direction        rotate;
      justification           justify;
      image_association       image;
      coordinate_pair         location;
      plot_text *             next;
   };
 
   plot_text * first_plot_text;
   plot_text * last_plot_text;
 
   void setup_plot_text_list()
   {
      first_plot_text        = NULL;
      last_plot_text         = NULL;
   };
 
   void kill_plot_text_list()
   {
      plot_text * condemned;
      plot_text * current = first_plot_text;
      while (current != NULL)
      {
         condemned = current;
         current = current->next;
         free(condemned->text_string);
         free(condemned->font_name);
         free(condemned);
      };
   };
 
   plot_text * set_plot_text(const char *       text_string,
                             const char *       font_name,
                             double             font_size,
                             rotate_direction   rotate,
                             justification      justify,
                             image_association  image,
                             coordinate_pair    location)
   {
      plot_text * new_plot_text;
      frail_allocate(new_plot_text, plot_text);
      new_plot_text->text_string     = frail_strdup(text_string);
      new_plot_text->font_name       = frail_strdup(font_name);
      new_plot_text->font_size       = font_size;
      new_plot_text->rotate          = rotate;
      new_plot_text->justify         = justify;
      new_plot_text->image           = image;
      new_plot_text->location        = location;
      new_plot_text->next            = NULL;
      if (last_plot_text == NULL)
      {
         first_plot_text = last_plot_text = new_plot_text;
      }
      else
      {
         last_plot_text->next = new_plot_text;
         last_plot_text = new_plot_text;
      };
      return new_plot_text;
   };
 
/* Print out stuff from the list of plot text */
 
   void emit_distinct_fonts(FILE * output_stream)
   {
      plot_text * current = first_plot_text;
      plot_text * previous;
      boolean duplicate;
 
      fprintf( output_stream, "%%%%DocumentFonts: " );
      while (current != NULL)
      {
         duplicate = LOGICAL_FALSE;
         previous = first_plot_text;
 
         while ((previous != current) && (duplicate == LOGICAL_FALSE))
         {
            if (strcmp(current->font_name, previous->font_name) == 0)
            {
               duplicate = LOGICAL_TRUE;
            };
            previous = previous->next;
         };
 
         if (duplicate == LOGICAL_FALSE)
         {
            fprintf(output_stream, " %s", current->font_name);
         };
         current  = current->next;
      };
      fprintf(output_stream, "\n");
   };
 
/* Print out all the text items associated with a particular image */
 
   void emit_plot_text(FILE * output_stream,
                       image_association image)
   {
      plot_text * current = first_plot_text;
      char dummy[] = "";
      char * current_font_name = dummy;
      double current_font_size = 0.0;
 
      fprintf(output_stream, "%% display titles\n" );
      while (current != NULL)
      {
         if (compatible_images(current->image, image) &&
             (current->text_string != NULL))
         {
            if ((strcmp(current->font_name, current_font_name) != 0) ||
                (current->font_size != current_font_size))
            {
               current_font_name = current->font_name;
               current_font_size = current->font_size;
               fprintf(output_stream, "/%s findfont %g scalefont setfont\n",
                                      current_font_name, current_font_size);
            };
            emit_coordinate_pair(output_stream, current->location, "moveto");
            if (current->rotate == ROTATE_NINETY)
            {
               fprintf(output_stream, " gsave 90 rotate" );
            };
            if (current->rotate == ROTATE_MINUS_NINETY)
            {
               fprintf(output_stream, " gsave -90 rotate" );
            };
            switch (current->justify)
            {
               case LEFT_JUSTIFY:
               {
                  fprintf(output_stream, " (%s) rshow", current->text_string);
               }; break;
               case RIGHT_JUSTIFY:
               {
                  fprintf(output_stream, " (%s) show", current->text_string);
               }; break;
               default:
               {
                  fprintf(output_stream, " (%s) cshow", current->text_string);
               };
            }
            if (current->rotate != DO_NOT_ROTATE)
            {
               fprintf(output_stream, " grestore\n");
            }
            else
            {
               fprintf(output_stream, "\n");
            };
         };
         current  = current->next;
      };
   };
 
 
/* Some stuff interweaving scales, text, and lines */
   char * make_label(UInt64 value, value_format vf)
   {
      static char temp[80];
      switch (vf)
      {
         case HEXADECIMAL_NUMBERS:
         {
            sprintf(temp, "%X hex", value);
         }; break;
         case DECIMAL_NUMBERS:
         {
            sprintf(temp, "%d", value);
         }; break;
         default:
         {
            sprintf(temp, "%X (%d)", value, value);
         };
      };
      return frail_strdup(temp);
   };
 
   double maximum_label_size(scale_info si)
   {
      char * temp = make_label(si.user_max, si.style);
      double label_size = strlen(temp) * si.font_size
                            * FONT_WIDTH_RATIO_ESTIMATE;
      free(temp);
      return label_size;
   };
 
/* Use the scale to emit tics and labels, etc. */
 
private:
   UInt32 maximum_major_tics(scale_info si)
   {
      if (si.labels_requested == LOGICAL_FALSE)
      {
         return MAXIMUM_MAJOR_TICS;
      }
      else
      {
         double available_length = fabs((position(* si.box_loc_max_ptr)
                                         - position(* si.box_loc_min_ptr))
                                   * si.storage_max / (si.storage_max + 1.0));
         UInt32 temp = (UInt32)(available_length / (2.5 * si.font_size));
         if (temp == 0)
         {
            return 1;
         }
         else
         {
            return temp;
         };
      };
   };
 
   void calculate_dec_tic_spacing(scale_info & si)
   {
      UInt64 range = si.user_max - si.user_min;
      UInt32 max_major_tics = maximum_major_tics(si);
 
      si.major_spacing      = 1;
      while (si.major_spacing * max_major_tics < range)
      {
         if (si.major_spacing * 2 * max_major_tics > range)
         {
            si.major_spacing     *= 2;
            si.minor_spacing      = si.major_spacing /  4;
            si.very_minor_spacing = si.major_spacing / 20;
         }
         else if (si.major_spacing * 4 * max_major_tics > range)
         {
            si.major_spacing     *= 4;
            si.minor_spacing      = si.major_spacing / 4;
            si.very_minor_spacing = si.major_spacing / 20;
         }
         else if (si.major_spacing * 5 * max_major_tics > range)
         {
            si.major_spacing     *= 5;
            si.minor_spacing      = si.major_spacing / 5;
            si.very_minor_spacing = si.major_spacing / 25;
         }
         else if (si.major_spacing * 10 * max_major_tics > range)
         {
            si.major_spacing     *= 10;
            si.minor_spacing      = si.major_spacing / 2;
            si.very_minor_spacing = si.major_spacing / 10;
         }
         else if (si.major_spacing * 20 * max_major_tics > range)
         {
            si.major_spacing     *= 20;
            si.minor_spacing      = si.major_spacing / 4;
            si.very_minor_spacing = si.major_spacing / 20;
         }
         else if (si.major_spacing * 25 * max_major_tics > range)
         {
            si.major_spacing     *= 25;
            si.minor_spacing      = si.major_spacing / 5;
            si.very_minor_spacing = si.major_spacing / 25;
         }
         else if (si.major_spacing * 40 * max_major_tics > range)
         {
            si.major_spacing     *= 40;
            si.minor_spacing      = si.major_spacing / 4;
            si.very_minor_spacing = si.major_spacing / 20;
         }
         else if (si.major_spacing * 50 * max_major_tics > range)
         {
            si.major_spacing     *= 50;
            si.minor_spacing      = si.major_spacing / 5;
            si.very_minor_spacing = si.major_spacing / 25;
         }
         else
         {
            si.major_spacing     *= 100;
            si.minor_spacing      = si.major_spacing / 4;
            si.very_minor_spacing = si.major_spacing / 10;
         };
      };
      if (si.major_spacing < 10)
      {
         si.minor_tics_requested = LOGICAL_FALSE;
         si.minor_spacing        = 1;
         si.very_minor_spacing   = 1;
      };
   };
 
   void calculate_hex_tic_spacing(scale_info & si)
   {
      UInt64 range = si.user_max - si.user_min;
      UInt32 max_major_tics = maximum_major_tics(si);
 
      si.major_spacing      = 1;
      while (si.major_spacing * max_major_tics < range)
      {
         si.major_spacing     *= 4;
      };
      if (si.major_spacing < 16)
      {
         si.minor_tics_requested = LOGICAL_FALSE;
         si.minor_spacing        = 1;
         si.very_minor_spacing   = 1;
      }
      else
      {
         si.minor_spacing      = si.major_spacing / 4;
         si.very_minor_spacing = si.major_spacing / 16;
      };
   };
 
   void calculate_first_tics_and_spacing(scale_info & si)
   {
      if (si.style == DECIMAL_NUMBERS)
      {
         calculate_dec_tic_spacing(si);
      }
      else
      {
         calculate_hex_tic_spacing(si);
      };
      si.first_major_tic      = (UInt64)(ceil(((double)si.user_min)/
                                              ((double)si.major_spacing))
                                         * si.major_spacing);
      si.first_minor_tic      = (UInt64)(ceil(((double)si.user_min)/
                                              ((double)si.minor_spacing))
                                         * si.minor_spacing);
      si.first_very_minor_tic = (UInt64)(ceil(((double)si.user_min)/
                                              ((double)si.very_minor_spacing))
                                         * si.very_minor_spacing);
   };
 
   void emit_line(FILE * output_stream,
                 double start_x_position,
                 double start_y_position,
                 double finish_x_position,
                 double finish_y_position)
   {
      fprintf(output_stream,
              "        %g %g moveto %g %g lineto\n",
              start_x_position,
              start_y_position,
              finish_x_position,
              finish_y_position);
   };
 
   void emit_vertical_tics(FILE * output_stream, scale_info & si)
   {
      double x_axis_center = position(si.axis_center);
      double x_tic_start;
      double x_tic_end;
      double y_pos;
      UInt64 loop;
      if (si.minor_tics_requested == LOGICAL_TRUE)
      {
         x_tic_start = x_axis_center - 0.5 * VERY_MINOR_TIC_SIZE;
         x_tic_end   = x_axis_center + 0.5 * VERY_MINOR_TIC_SIZE;
         fprintf(output_stream, "newpath %%very_minor tics \n");
         for (loop  = si.first_very_minor_tic;
              loop <= si.user_max;
              loop += si.very_minor_spacing)
         {
            if ((loop % si.minor_spacing) != 0)
            {
               y_pos = position(box_center(si, loop));
               emit_line(output_stream,
                         x_tic_start, y_pos,
                         x_tic_end,   y_pos);
            };
         };
         fprintf(output_stream,
                 "closepath %g setlinewidth 1 setlinecap stroke\n",
                 VERY_MINOR_TIC_THICKNESS);

         x_tic_start = x_axis_center - 0.5 * MINOR_TIC_SIZE;
         x_tic_end   = x_axis_center + 0.5 * MINOR_TIC_SIZE;
         fprintf(output_stream, "newpath %%minor tics \n");
         for (loop  = si.first_minor_tic;
              loop <= si.user_max;
              loop += si.minor_spacing)
         {
            if ((loop % si.major_spacing) != 0)
            {
               y_pos = position(box_center(si, loop));
               emit_line(output_stream,
                         x_tic_start, y_pos,
                         x_tic_end,   y_pos);
            };
         };
         fprintf(output_stream,
                 "closepath %g setlinewidth 1 setlinecap stroke\n",
                 MINOR_TIC_THICKNESS);
      };
 
      x_tic_start = x_axis_center - 0.5 * MAJOR_TIC_SIZE;
      x_tic_end   = x_axis_center + 0.5 * MAJOR_TIC_SIZE;
      fprintf(output_stream, "newpath %%major tics \n");
      for (loop  = si.first_major_tic;
           loop <= si.user_max;
           loop += si.major_spacing)
      {
         y_pos = position(box_center(si, loop));
         emit_line(output_stream, x_tic_start, y_pos, x_tic_end, y_pos);
      };
      fprintf(output_stream,
              "closepath %g setlinewidth 1 setlinecap stroke\n",
              MAJOR_TIC_THICKNESS);
   };
 
   void emit_horizontal_tics(FILE * output_stream, scale_info si)
   {
      double y_axis_center = position(si.axis_center);
      double y_tic_start;
      double y_tic_end;
      double x_pos;
      UInt64 loop;
      if (si.minor_tics_requested == LOGICAL_TRUE)
      {
         y_tic_start = y_axis_center - 0.5 * VERY_MINOR_TIC_SIZE;
         y_tic_end   = y_axis_center + 0.5 * VERY_MINOR_TIC_SIZE;
         fprintf(output_stream, "newpath %%very_minor tics \n");
         for (loop  = si.first_very_minor_tic;
              loop <= si.user_max;
              loop += si.very_minor_spacing)
         {
            if ((loop % si.minor_spacing) != 0)
            {
               x_pos = position(box_center(si, loop));
               emit_line(output_stream,
                         x_pos, y_tic_start,
                         x_pos, y_tic_end);
            };
         };
         fprintf(output_stream,
                 "closepath %g setlinewidth 1 setlinecap stroke\n",
                 VERY_MINOR_TIC_THICKNESS);
 
         y_tic_start = y_axis_center - 0.5 * MINOR_TIC_SIZE;
         y_tic_end   = y_axis_center + 0.5 * MINOR_TIC_SIZE;
         fprintf(output_stream, "newpath %%minor tics \n");
         for (loop  = si.first_minor_tic;
              loop <= si.user_max;
              loop += si.minor_spacing)
         {
            if ((loop % si.major_spacing) != 0)
            {
               x_pos = position(box_center(si, loop));
               emit_line(output_stream,
                         x_pos, y_tic_start,
                         x_pos, y_tic_end);
            };
         };
         fprintf(output_stream,
                 "closepath %g setlinewidth 1 setlinecap stroke\n",
                 MINOR_TIC_THICKNESS);
      };
 
      y_tic_start = y_axis_center - 0.5 * MAJOR_TIC_SIZE;
      y_tic_end   = y_axis_center + 0.5 * MAJOR_TIC_SIZE;
      fprintf(output_stream, "newpath %%major tics \n");
      for (loop  = si.first_major_tic;
           loop <= si.user_max;
           loop += si.major_spacing)
      {
         x_pos = position(box_center(si, loop));
         emit_line(output_stream, x_pos, y_tic_start, x_pos, y_tic_end);
      };
      fprintf(output_stream,
              "closepath %g setlinewidth 1 setlinecap stroke\n",
              MAJOR_TIC_THICKNESS);
   };

   void emit_vertical_labels(FILE * output_stream, scale_info si)
   {
      double x_labels_left_edge = position(si.label_edge);
      double y_pos;

      fprintf(output_stream, "%%vertical tic labels \n");
      fprintf(output_stream, "gsave /%s findfont %g scalefont setfont\n",
                             si.font_name, si.font_size);
      for (UInt64 loop  = si.first_major_tic;
                  loop <= si.user_max;
                  loop += si.major_spacing)
      {
         y_pos = position(box_center(si, loop)) - 0.33 * si.font_size;
         fprintf(output_stream, "      %g %g moveto (%s) rshow\n",
                                x_labels_left_edge, y_pos,
                                make_label(loop, si.style));
      };
      fprintf(output_stream, "grestore\n");
   };

   void emit_horizontal_labels(FILE * output_stream, scale_info si)
   {
      double y_labels_top_edge = position(si.label_edge);
      double x_pos;

      fprintf(output_stream, "%%horizontal tic labels \n");
      fprintf(output_stream, "gsave -90 rotate \n");
      fprintf(output_stream, "      /%s findfont %g scalefont setfont\n",
                             si.font_name, si.font_size);
      for (UInt64 loop  = si.first_major_tic;
                  loop <= si.user_max;
                  loop += si.major_spacing)
      {
         x_pos = position(box_center(si, loop)) - 0.33 * si.font_size;
         fprintf(output_stream, "      %g %g moveto (%s) show \n",
                                - y_labels_top_edge, x_pos,
                                make_label(loop, si.style));
      };
      fprintf(output_stream, "grestore\n");
   };

protected:
   void emit_vertical_scale(FILE * output_stream, scale_info & si)
   {
      calculate_first_tics_and_spacing(si);
      if (si.tics_requested == LOGICAL_TRUE)
      {
         fprintf(output_stream, "newpath %%center line \n");
         double x_pos        = position(si.axis_center);
         double y_pos_start  = position(box_center(si, si.user_min));
         double y_pos_finish = position(box_center(si, si.user_max));
         emit_line(output_stream, x_pos, y_pos_start,
                                  x_pos, y_pos_finish);
         fprintf(output_stream,
              "closepath %g setlinewidth 1 setlinecap stroke\n",
              MAJOR_TIC_THICKNESS);
         emit_vertical_tics(output_stream, si);
      };
      if (si.labels_requested == LOGICAL_TRUE)
      {
         emit_vertical_labels(output_stream, si);
      }
   };
 
   void emit_horizontal_scale(FILE * output_stream, scale_info & si)
   {
      calculate_first_tics_and_spacing(si);
      if (si.tics_requested == LOGICAL_TRUE)
      {
         fprintf(output_stream, "newpath %%center line \n");
         double x_pos_start  = position(box_center(si, si.user_min));
         double x_pos_finish = position(box_center(si, si.user_max));
         double y_pos        = position(si.axis_center);
         emit_line(output_stream, x_pos_start, y_pos,
                                  x_pos_finish, y_pos);
         fprintf(output_stream,
              "closepath %g setlinewidth 1 setlinecap stroke\n",
              MAJOR_TIC_THICKNESS);
         emit_horizontal_tics(output_stream, si);
      };
      if (si.labels_requested == LOGICAL_TRUE)
      {
         emit_horizontal_labels(output_stream, si);
      }
   };
 
   plot_basics()
   {
      setup_plot_text_list();
   };
 
   void kill_plot_basics()
   {
      kill_plot_text_list();
   };
 
   friend class standard_plot;
   friend class density_plot;
   friend class histogram_plot;
};
#endif

